import json
import os
from datetime import datetime
from decimal import Decimal
from unittest import mock


from satcfdi import Signer
from satcfdi.create.cfd import cfdi40
from satcfdi.pacs import Environment
from satcfdi.pacs.mysuite import MYSuite
from utils import get_signer, verify_result

current_dir = os.path.dirname(__file__)


def test_mysuite_test():
    mysuite = MYSuite(
        requestor="0c320b03-d4f1-47bc-9fb4-77995f9bf33e",
        country="MX",
        user_name="Juan",
        environment=Environment.TEST
    )

    signer = Signer.load(
        certificate=open(os.path.join(current_dir, f'csd/Certificado_Prueba_Vigente_SAT_JES900109Q90/30001000000400002436.cer'), 'rb').read(),
        key=open(os.path.join(current_dir, f'csd/Certificado_Prueba_Vigente_SAT_JES900109Q90/30001000000400002436.key'), 'rb').read(),
        password=open(os.path.join(current_dir, f'csd/Certificado_Prueba_Vigente_SAT_JES900109Q90/pass.txt'), 'rb').read()
    )
    emisor = cfdi40.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )

    invoice = cfdi40.Comprobante(
        emisor=emisor,
        lugar_expedicion="27200",
        fecha=datetime.fromisoformat("2022-09-28T22:40:38"),
        receptor=cfdi40.Receptor(
            rfc="JES900109Q90",
            nombre="JIMENEZ ESTRADA SALAS A A",
            uso_cfdi="G01",
            domicilio_fiscal_receptor="01030",
            regimen_fiscal_receptor="601"
        ),
        metodo_pago="PPD",
        forma_pago="99",
        serie="T",
        folio="1000",
        conceptos=cfdi40.Concepto(
            cuenta_predial='1234567890',
            clave_prod_serv='10101702',
            cantidad=Decimal('.10'),
            clave_unidad='E48',
            descripcion='SERVICIOS DE RENTA',
            valor_unitario=Decimal("1.00"),
            impuestos=cfdi40.Impuestos(
                traslados=cfdi40.Impuesto.parse('002|Tasa|0.160000'),
                retenciones=[cfdi40.Impuesto.parse('001|Tasa|0.100000'), cfdi40.Impuesto.parse('002|Tasa|0.106667')],
            ),
            _traslados_incluidos=True
        )
    )
    invoice.sign(signer)

    with mock.patch(f'requests.post') as mk:
        mk.return_value.content = b"""<?xml version="1.0" encoding="utf-8"?>
        <soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema">
            <soap:Body>
                <RequestTransactionResponse xmlns="http://www.fact.com.mx/schema/ws">
                    <RequestTransactionResult>
                        <Request>
                            <Requestor>00000000-0000-0000-0000-000000000000</Requestor>
                            <RequestorActive>true</RequestorActive>
                            <Transaction>TIMBRAR</Transaction>
                            <Country>MX</Country>
                            <Entity>JES900109Q90</Entity>
                            <User>00000000-0000-0000-0000-000000000000</User>
                            <UserName>MX.JES900109Q90.Juan</UserName>
                            <Id>96f342f4-a28a-487a-907f-7d00f388a8ce</Id>
                            <TimeStamp>2023-03-21T10:56:25.5256369-06:00</TimeStamp>
                        </Request>
                        <Response>
                            <Result>true</Result>
                            <TimeStamp>2023-03-21T10:56:26.1661906-06:00</TimeStamp>
                            <LastResult/>
                            <Code>1</Code>
                            <Description/>
                            <Hint/>
                            <Data> 20393 96f342f4-a28a-487a-907f-7d00f388a8ce</Data>
                            <Processor>TEST-BACK02</Processor>
                        </Response>
                        <ResponseData>
                            <ResponseData1>PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0iVVRGLTgiPz48dGZkOlRpbWJyZUZpc2NhbERpZ2l0YWwgeHNpOnNjaGVtYUxvY2F0aW9uPSJodHRwOi8vd3d3LnNhdC5nb2IubXgvVGltYnJlRmlzY2FsRGlnaXRhbCBodHRwOi8vd3d3LnNhdC5nb2IubXgvc2l0aW9faW50ZXJuZXQvY2ZkL1RpbWJyZUZpc2NhbERpZ2l0YWwvVGltYnJlRmlzY2FsRGlnaXRhbHYxMS54c2QiIFZlcnNpb249IjEuMSIgVVVJRD0iOTZmMzQyZjQtYTI4YS00ODdhLTkwN2YtN2QwMGYzODhhOGNlIiBGZWNoYVRpbWJyYWRvPSIyMDIzLTAzLTIxVDEwOjU2OjI2IiBSZmNQcm92Q2VydGlmPSJTUFIxOTA2MTNJNTIiIFNlbGxvQ0ZEPSJ5TU0rSVk3azd6UzloQzVvZkZGekM4bjVFendFL3Mvd3dWOE56M1l4dHh0K082Um1Cbm8vUTdENFJKeGpoS2dSOE1sRjJpQ0ROK1BkTmptWEZxYU04RnVnbkZOd2k1eEgxbVZsRk84OGcyT0REZXFib0d6bTFNN2lXM0IvN0RDTjRhRlZTa0hmdi9sUVFHNi9aWmFCMkdVNWhsMzd6MnVVRTZndW5Db1RtUUFQZUpOY1NxUjNJTTNyS3dPRmUxVGxDZm9iM3FBYWFwY0R4c0RYSStHM0huQ1llRW9JYjIyV1krRWNrTmhRdndlUUQydGtXTFZPc2xYNUxHTzVwSFRrR3dyZGV3L0NmVlVFeDVLNlpwK2lKaXFTQ0NqS1VLMythTVpLOURZeUdWTFRsQWdDR2FvZllZSDNJOGxFVFJUVXlmWHNIS0ZIa0xtaks4Q05RbjRJR0E9PSIgTm9DZXJ0aWZpY2Fkb1NBVD0iMzAwMDEwMDAwMDA0MDAwMDI0OTUiIFNlbGxvU0FUPSJJazd0dEpiYjcyNXJSNDFnUFYxQVJNdGdTc3UxT2FYRklyL3ZWaE5qU1ZtWVdlSHNYV2oyUHg3YzFhUW5YeXU4Uno0cXQxMGN2ZjMrQXZ4anVZcSszMmdsZmdFL0c2UHZWMVNYbkFKaHFoRk1sUFhJTFpHZ1Q5S0lXbE9ZdlVPRDJLMXgyTTBONzFlNWgwdW5HYVBHWTM3N3pWUlNDT20yR0U2VVBPZXhVQTB1OXRFSnIrQm5rR3dpZVpYbFJyY01DekowRTJwbVZBS2VFTndFRExOOWVGUjhGbno3M2hQZUtHTVFlNmc3VVZtczJvNktKbDlJMERuR1c4MGhoTWd4Uy9KdzM5RW5EazRtektVbVV1SHBqRnh5Rm1tZjlvY2FsOFlUMXVNTS95bjdhVHFwSFk4Vm8rZWZHb1VGWHRIWVllK3YzZVliOXdaZ0d1bE5hSHFhaGc9PSIgeG1sbnM6dGZkPSJodHRwOi8vd3d3LnNhdC5nb2IubXgvVGltYnJlRmlzY2FsRGlnaXRhbCIgeG1sbnM6eHNpPSJodHRwOi8vd3d3LnczLm9yZy8yMDAxL1hNTFNjaGVtYS1pbnN0YW5jZSIgLz4=</ResponseData1>
                            <ResponseData2>51749</ResponseData2>
                            <ResponseData3>OK</ResponseData3>
                        </ResponseData>
                    </RequestTransactionResult>
                </RequestTransactionResponse>
            </soap:Body>
        </soap:Envelope>
        """

        res = mysuite.stamp(
            cfdi=invoice
        )
        verify = verify_result(data=res.xml, filename="test_mysuite_stamp.xml")
        assert verify
        assert not res.pdf

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="test_mysuite_stamp.json")
        assert verify





from ...utils import StrEnum


class Retenciones(StrEnum):
    SERVICIOS_PROFESIONALES = '01'
    REGALIAS_POR_DERECHOS_DE_AUTOR = '02'
    AUTOTRANSPORTE_TERRESTRE_DE_CARGA = '03'
    SERVICIOS_PRESTADOS_POR_COMISIONISTAS = '04'
    ARRENDAMIENTO = '05'
    ENAJENACION_DE_ACCIONES = '06'
    ENAJENACION_DE_BIENES_OBJETO_DE_LA_LIEPS_A_TRAVES_DE_MEDIADORES_AGENTES_REPRESENTANTES_CORREDORES_CONSIGNATARIOS_O_DISTRIBUIDORES = '07'
    ENAJENACION_DE_BIENES_INMUEBLES_CONSIGNADA_EN_ESCRITURA_PUBLICA = '08'
    ENAJENACION_DE_OTROS_BIENES_NO_CONSIGNADA_EN_ESCRITURA_PUBLICA = '09'
    ADQUISICION_DE_DESPERDICIOS_INDUSTRIALES = '10'
    ADQUISICION_DE_BIENES_CONSIGNADA_EN_ESCRITURA_PUBLICA = '11'
    ADQUISICION_DE_OTROS_BIENES_NO_CONSIGNADA_EN_ESCRITURA_PUBLICA = '12'
    OTROS_RETIROS_DE_AFORE = '13'
    DIVIDENDOS_O_UTILIDADES_DISTRIBUIDAS = '14'
    REMANENTE_DISTRIBUIBLE = '15'
    INTERESES = '16'
    ARRENDAMIENTO_EN_FIDEICOMISO = '17'
    PAGOS_REALIZADOS_A_FAVOR_DE_RESIDENTES_EN_EL_EXTRANJERO = '18'
    ENAJENACION_DE_ACCIONES_U_OPERACIONES_EN_BOLSA_DE_VALORES = '19'
    OBTENCION_DE_PREMIOS = '20'
    FIDEICOMISOS_QUE_NO_REALIZAN_ACTIVIDADES_EMPRESARIALES = '21'
    PLANES_PERSONALES_DE_RETIRO = '22'
    INTERESES_REALES_DEDUCIBLES_POR_CREDITOS_HIPOTECARIOS = '23'
    OPERACIONES_FINANCIERAS_DERIVADAS_DE_CAPITAL = '24'
    OTRO_TIPO_DE_RETENCIONES = '25'
    SERVICIOS_MEDIANTE_PLATAFORMAS_TECNOLOGICAS = '26'


class TipoContribuyenteSujetoRetencion(StrEnum):
    ARTISTAS_DEPORTISTAS_Y_ESPECTACULOS_PUBLICOS = '1'
    OTRAS_PERSONAS_FISICAS = '2'
    PERSONA_MORAL = '3'
    FIDEICOMISO = '4'
    ASOCIACION_EN_PARTICIPACION = '5'
    ORGANIZACIONES_INTERNACIONALES_O_DE_GOBIERNO = '6'
    ORGANIZACIONES_EXENTAS = '7'
    AGENTES_PAGADORES = '8'
    OTROS = '9'


class TipoDividendoOUtilidadDistribuida(StrEnum):
    PROVIENE_DE_CUFIN = '01'
    NO_PROVIENE_DE_CUFIN = '02'
    REEMBOLSO_O_REDUCCION_DE_CAPITAL = '03'
    LIQUIDACION_DE_LA_PERSONA_MORAL = '04'
    CUFINRE = '05'
    PROVIENE_DE_CUFIN_AL_31_DE_DICIEMBRE_2013 = '06'


class TipoImpuesto(StrEnum):
    ISR = '01'
    IVA = '02'
    IEPS = '03'


class TipoPagoRet(StrEnum):
    PAGO_DEFINITIVO_IVA = '01'
    PAGO_DEFINITIVO_IEPS = '02'
    PAGO_DEFINITIVO_ISR = '03'
    PAGO_PROVISIONAL_ISR = '04'


class TipoAportODep(StrEnum):
    APORTACIONES_VOLUNTARIAS_O_COMPLEMENTARIAS_EFECTUADAS_EN_TERMINOS_DEL_ARTICULO_151_FRACCION_V_DE_LA_LISR = 'a'
    DEPOSITOS_A_CUENTAS_ESPECIALES_PRIMA_DE_SEGUROS_O_ADQUISICION_DE_ACCIONES_EFECTUADAS_EN_TERMINOS_AL_ARTICULO_185_DE_LA_LISR = 'b'
    APORTACIONES_ADICIONALES_EFECTUADAS_EN_TERMINOS_DEL_ARTICULO_258 = 'c'





# !/usr/bin/env python3

"""Implements the client for Series."""

# pylint: skip-file
# pydocstyle: add-ignore=D105,D107,D401

from typing import Dict, Optional

import requests
import requests.auth

from .. import __version__, ResponseError


class SieAPIRest:
    """
    SIE-BANXICO API
    API de consulta de series de tiempo del Banco de México
    version: 1.4.3-hotfix1
    """
    def __init__(
            self,
            bmx_token: str,
            url_prefix: str = "https://www.banxico.org.mx") -> None:
        self.headers = {
            'Bmx-Token': bmx_token,
            'Accept': 'application/json',
            "User-Agent": __version__.__user_agent__
        }
        self.url_prefix = url_prefix

    def _request(self, method, path, params=None):
        print(f"{self.url_prefix}/{path}")

        resp = requests.request(
            method=method,
            url=f"{self.url_prefix}/{path}",
            headers=self.headers,
            params=params
        )

        if resp.status_code == 200:
            return resp.json()

        raise ResponseError(resp.text)

    def metadatos_series_using_get(
            self,
            id_series: str,
    ) -> str:
        """
        Series metadata
        Send a get request to SieAPIRest/service/v1/series/{idSeries}.

        :param id_series: idSeries
        :return: OK
        """
        return self._request(
            method="get",
            path=f"SieAPIRest/service/v1/series/{id_series}"
        )

    def datos_series_using_get(
            self,
            id_series: str,
            decimales: Optional[str] = None,
            incremento: Optional[str] = None,
    ) -> str:
        """
        Series data
        Send a get request to SieAPIRest/service/v1/series/{idSeries}/datos.

        :param id_series: idSeries
        :param decimales: 'sinCeros' to remove trailing zeros to the right of the decimal point.
        :param incremento: 'PorcObsAnt', 'PorcAnual', 'PorcAcumAnual'.
        :return: OK
        """
        params = {}  # type: Dict[str, str]
        if decimales is not None:
            params['decimales'] = decimales
        if incremento is not None:
            params['incremento'] = incremento

        return self._request(
            method='get',
            path=f"SieAPIRest/service/v1/series/{id_series}/datos",
            params=params
        )

    def datos_oportuno_series_using_get(
            self,
            id_series: str,
            decimales: Optional[str] = None,
            incremento: Optional[str] = None,
    ) -> str:
        """
        Last data
        Send a get request to SieAPIRest/service/v1/series/{idSeries}/datos/oportuno.

        :param id_series: idSeries
        :param decimales: 'sinCeros' to remove trailing zeros to the right of the decimal point.
        :param incremento: 'PorcObsAnt', 'PorcAnual', 'PorcAcumAnual'.
        :return: OK
        """
        params = {}  # type: Dict[str, str]
        if decimales is not None:
            params['decimales'] = decimales
        if incremento is not None:
            params['incremento'] = incremento

        return self._request(
            method='get',
            path=f"SieAPIRest/service/v1/series/{id_series}/datos/oportuno",
            params=params,
        )

    def obten_rango_datos_series_using_get(
            self,
            id_series: str,
            fecha_inicial: str,
            fecha_final: str,
            decimales: Optional[str] = None,
            incremento: Optional[str] = None,
    ) -> str:
        """
        Data range
        Send a get request to SieAPIRest/service/v1/series/{idSeries}/datos/{fechaInicial}/{fechaFinal}.

        :param id_series: idSeries
        :param fecha_inicial: fechaInicial
        :param fecha_final: fechaFinal
        :param decimales: 'sinCeros' to remove trailing zeros to the right of the decimal point.
        :param incremento: 'PorcObsAnt', 'PorcAnual', 'PorcAcumAnual'.
        :return: OK
        """
        params = {}  # type: Dict[str, str]
        if decimales is not None:
            params['decimales'] = decimales
        if incremento is not None:
            params['incremento'] = incremento

        return self._request(
            method='get',
            path=f"SieAPIRest/service/v1/series/{id_series}/datos/{fecha_inicial}/{fecha_final}",
            params=params,
        )

    def metadatos_series_versionadas_using_get(
            self,
            id_series: str,
    ) -> str:
        """
        Series versions metadata
        Send a get request to SieAPIRest/service/v1/series/{idSeries}/versiones.

        :param id_series: idSeries
        :return: OK
        """
        return self._request(
            method='get',
            path=f"SieAPIRest/service/v1/series/{id_series}/versiones",
        )

    def datos_series_versionadas_using_get(
            self,
            id_series: str,
            id_version: str,
            decimales: Optional[str] = None,
            incremento: Optional[str] = None,
    ) -> str:
        """
        Series data
        Send a get request to SieAPIRest/service/v1/series/{idSeries}/versiones/{idVersion}/datos.

        :param id_series: idSeries
        :param id_version: idVersion
        :param decimales: 'sinCeros' to remove trailing zeros to the right of the decimal point.
        :param incremento: 'PorcObsAnt', 'PorcAnual', 'PorcAcumAnual'.
        :return: OK
        """
        params = {}  # type: Dict[str, str]
        if decimales is not None:
            params['decimales'] = decimales
        if incremento is not None:
            params['incremento'] = incremento

        return self._request(
            method='get',
            path=f"SieAPIRest/service/v1/series/{id_series}/versiones/{id_version}/datos",
            params=params,
        )



from datetime import datetime
import requests
import urllib3
from bs4 import BeautifulSoup
# noinspection PyUnresolvedReferences
from ..transform.catalog import CATALOGS
from .. import ResponseError, __version__, Code

try:
    urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH'
except:
    pass


def retrieve(rfc: str, id_cif: str):
    data = _request_constancia(rfc, id_cif)
    return _parse_response(data)


def url(rfc: str, id_cif: str):
    return f"https://siat.sat.gob.mx/app/qr/faces/pages/mobile/validadorqr.jsf?D1=10&D2=1&D3={id_cif}_{rfc}"


def _request_constancia(rfc: str, id_cif: str):
    res = requests.get(
        url=url(rfc, id_cif),
        headers={
            "User-Agent": __version__.__user_agent__,
        }
    )
    if res.ok:
        return res.content
    else:
        raise ResponseError(res)


def _find_regimen(regimen):
    for k, v in CATALOGS['{http://www.sat.gob.mx/sitio_internet/cfd/catalogos}c_RegimenFiscal'].items():
        if regimen.endswith(v):
            return Code(k, v)
    return Code(None, regimen)


def _parse_response(data):
    _REGIMENES = "Regimenes"
    html = BeautifulSoup(data, 'html.parser')
    gc_v = html.find_all(name="td", attrs={"role": "gridcell", "style": "text-align:left;"})
    if not gc_v:
        raise ValueError("'rfc' or 'id_cif' is invalid")
    gc_k = html.find_all(name="span", attrs={"style": "font-weight: bold;"})
    result = {_REGIMENES: []}

    for k, v in zip(gc_k, gc_v):
        k = k.text.rstrip(":")
        v = v.text
        if k.startswith("Fecha ") and v:
            v = datetime.strptime(v, "%d-%m-%Y").date()

        if k == 'Régimen':
            result[_REGIMENES].append({
                'RegimenFiscal': _find_regimen(v),
            })
        elif k == 'Fecha de alta':
            result[_REGIMENES][-1][k] = v
        else:
            result[k] = v

    return result



import os.path
from datetime import datetime, timedelta
from decimal import Decimal
from pprint import PrettyPrinter
from unittest import mock

from satcfdi import __version__, CFDI, Code
from satcfdi.pacs import Environment
from satcfdi.pacs import TaxpayerStatus
from satcfdi.pacs.sat import _CFDISolicitaDescarga, _CFDIAutenticacion, EstadoSolicitud, TipoDescargaMasivaTerceros
from satcfdi.pacs.sat import SAT
from satcfdi.pacs.sat import _get_listado_69b

from tests.utils import get_signer, verify_result

module = 'satcfdi'
current_dir = os.path.dirname(__file__)
pp = PrettyPrinter()


def test_pac_sat():
    sat_service = SAT(environment=Environment.TEST)

    with mock.patch(f'requests.post') as mk:
        mk.return_value.ok = True
        mk.return_value.content = (
            b'<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"><s:Body><ConsultaResponse xmlns="http://tempuri.org/"><ConsultaResult '
            b'xmlns:a="http://schemas.datacontract.org/2004/07/Sat.Cfdi.Negocio.ConsultaCfdi.Servicio" xmlns:i="http://www.w3.org/2001/XMLSchema-instance"><a:CodigoEstatus>S - '
            b'Comprobante obtenido satisfactoriamente.</a:CodigoEstatus><a:EsCancelable>No '
            b'cancelable</a:EsCancelable><a:Estado>Vigente</a:Estado><a:EstatusCancelacion/><a:ValidacionEFOS>200</a:ValidacionEFOS></ConsultaResult></ConsultaResponse></s:Body'
            b'></s:Envelope> '
        )

        cfdi = CFDI({
            "Emisor": {
                "Rfc": 'RFC_EMISOR'
            },
            "Receptor": {
                "Rfc": 'RFC_RECEPTOR'
            },
            "Total": Decimal('123456.00'),
            "Complemento": {
                "TimbreFiscalDigital": {
                    "UUID": "6114cfd0-87d2-45b8-99b1-7c19475c9cda"
                }
            }
        })
        # res = sat_service.consulta(
        #     rfc_emisor='RFC_EMISOR',
        #     rfc_receptor='RFC_RECEPTOR',
        #     total='123,456.00',
        #     uuid="6114cfd0-87d2-45b8-99b1-7c19475c9cda"
        # )
        res = sat_service.status(
            cfdi
        )

        assert mk.called
        print(mk.call_args.kwargs)
        assert mk.call_args.kwargs == {
            'url': 'https://pruebacfdiconsultaqr.cloudapp.net/ConsultaCFDIService.svc',
            'data': '<Envelope xmlns="http://schemas.xmlsoap.org/soap/envelope/" xmlns:tem="http://tempuri.org/"><Body><tem:Consulta><tem:expresionImpresa><![CDATA[?re=RFC_EMISOR&rr=RFC_RECEPTOR&tt=123456.00&id=6114cfd0-87d2-45b8-99b1-7c19475c9cda]]></tem:expresionImpresa></tem:Consulta></Body></Envelope>',
            'headers': {
                'Content-type': 'text/xml;charset="utf-8"',
                'Accept': 'text/xml',
                'Cache-Control': 'no-cache',
                'SOAPAction': 'http://tempuri.org/IConsultaCFDIService/Consulta',
                'User-Agent': __version__.__user_agent__,
            },
            'verify': False
        }

    assert res == {'CodigoEstatus': 'S - Comprobante obtenido satisfactoriamente.', 'EsCancelable': 'No cancelable', 'Estado': 'Vigente', 'EstatusCancelacion': None,
                   'ValidacionEFOS': '200'}


def test_listado69b():
    listado = _get_listado_69b()
    assert "SAT970701NN3" not in listado
    assert len(listado) > 12000
    values = [e.value for e in TaxpayerStatus]

    assert all(x in values for x in listado.values())
    assert listado['AAL081211JP0'] == TaxpayerStatus.DEFINITIVO.value


def test_sat_list_69b():
    sat_service = SAT(environment=Environment.TEST)
    res = sat_service.list_69b('AAL081211JP0')
    assert res == TaxpayerStatus.DEFINITIVO


def test_sat_list_69b_notfound():
    sat_service = SAT(environment=Environment.TEST)
    res = sat_service.list_69b('xxxx')
    assert res is None


def test_sat_service_authentication():
    signer = get_signer('xiqb891116qe4')

    with mock.patch(f'{module}.pacs.sat.datetime') as m:
        m.utcnow = mock.Mock(return_value=datetime(2022, 1, 1))

        req = _CFDIAutenticacion(signer=signer)
        res = req.get_payload()

        verify = verify_result(data=res, filename="test_sat_service_authentication.xml")
        assert verify


def test_sat_service_solicitud():
    signer = get_signer('xiqb891116qe4')

    with mock.patch(f'{module}.pacs.sat.datetime') as m:
        m.utcnow = mock.Mock(return_value=datetime(2022, 1, 1))

        req = _CFDISolicitaDescarga(
            signer=signer,
            arguments={
                'FechaFinal': 'FechaFinal',
                'FechaInicial': 'FechaInicial',
                'RfcEmisor': "RfcEmisor",
                'RfcSolicitante': "RfcSolicitante",
                'TipoSolicitud': "TipoSolicitud",
                'TipoComprobante': "TipoComprobante",
                'EstadoComprobante': "EstadoComprobante",
                'RfcACuentaTerceros': "RfcACuentaTerceros",
                'Complemento': "Complemento",
                'UUID': "UUID",
            }

        )
        res = req.get_payload()

        verify = verify_result(data=res, filename="test_sat_service_solicitud.xml")
        assert verify


def test_status_code():
    assert EstadoSolicitud(5).name == 'RECHAZADA'

    est_code = Code(5, EstadoSolicitud(5).name)
    alt_text = str(est_code)
    assert alt_text == "5 - RECHAZADA"
    assert EstadoSolicitud.RECHAZADA == est_code
    assert est_code == EstadoSolicitud.RECHAZADA


def test_pac_sat_uuid():
    signer = get_signer('xiqb891116qe4')
    sat_service = SAT(environment=Environment.TEST, signer=signer)

    sat_service.token_comprobante = {
        "Expires": datetime.utcnow() + timedelta(seconds=3600),
        "AutenticaResult": "token_comprobante"
    }

    with open(os.path.join(current_dir, 'pac_sat_responses', 'response-with-id.xml'), 'rb') as f:
        content = f.read()

    with mock.patch(f'requests.post') as mk:
        mk.return_value.ok = True
        mk.return_value.content = content

        res = sat_service.recover_comprobante_request(
            tipo_solicitud=TipoDescargaMasivaTerceros.CFDI,
            uuid="6114cfd0-87d2-45b8-99b1-7c19475c9cda",
        )

        mk.assert_called_once()
        del mk.call_args.kwargs['headers']['User-Agent']

        verify = verify_result(data=pp.pformat(mk.call_args.kwargs), filename=f"pac_sat_uuid.pretty.py")
        assert verify




import filecmp
import glob
import json
import os
from unittest import mock

import xlsxwriter
from lxml.etree import QName
from satcfdi.printer import Representable

from satcfdi.transform.pdf_environment import PDFEnvironment

from satcfdi import DatePeriod
from satcfdi.accounting._ansi_colors import *
from satcfdi.accounting.formatters import SatCFDI
from satcfdi.accounting.process import filter_invoices_iter, invoices_export, invoices_print, payments_print, \
    complement_invoices_data, payments_export, num2col, filter_payments_iter, payments_retentions_export, filter_retenciones_iter, retenciones_print, payments_groupby_receptor
from tests.utils import verify_result

current_dir = os.path.abspath(os.path.dirname(__file__))


def test_excel_utils():
    assert num2col(0) == "A"
    assert num2col(1) == "B"
    assert num2col(28) == "AC"


def test_ansi():
    assert paint("HOla", COLOR_BACKGROUND_BLACK) == '\x1b[40mHOla\x1b[39m'


class myCFDI(SatCFDI):
    @SatCFDI.estatus.getter
    def estatus(self) -> str:
        return '1'

    def consulta_estado(self):
        return {}


def test_cfdi():
    rfc = "H&E951128469"
    dp = DatePeriod(2020)

    all_invoices = {}

    for f in glob.iglob(os.path.join(current_dir, "invoices", "*.xml")):
        c = myCFDI.from_file(f)
        all_invoices[c.uuid] = c

    complement_invoices_data(all_invoices)

    for c in all_invoices.values():
        d = c.ultima_num_parcialidad
        assert d >= 0

    with mock.patch('builtins.print') as p:
        ingresos_pendientes = list(filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_emisor=rfc, invoice_type="I", pending_balance=lambda x: x > 0))
        invoices_print(ingresos_pendientes)

        pagos_pendientes = list(filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_receptor=rfc, invoice_type="I", pending_balance=lambda x: x > 0))
        invoices_print(pagos_pendientes)

        emitidas = list(filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_emisor=rfc))
        invoices_print(emitidas)

        pagos = list(filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_emisor=rfc))
        payments_print(pagos)

        recibidas = list(filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_receptor=rfc))
        invoices_print(recibidas)

        pagos_hechos = list(filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_receptor=rfc))
        payments_print(pagos_hechos)

        path = os.path.join(current_dir, "test_accounting", "invoices.xlsx")

        workbook = xlsxwriter.Workbook(path)
        invoices_export(workbook, "EMITIDAS", emitidas)
        payments_export(workbook, "EMITIDAS PAGOS", pagos)
        invoices_export(workbook, "RECIBIDAS", recibidas)
        payments_export(workbook, "PAGOS EMITIDOS", pagos_hechos)
        workbook.close()

        pagos = list(filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_emisor=rfc))
        pagos_agrupados = payments_groupby_receptor(pagos)
        payments_retentions_export(os.path.join(current_dir, "test_accounting", "retenciones.txt"), pagos_agrupados)


def test_retenciones():
    year = 2020
    all_invoices = {}

    for f in glob.iglob(os.path.join(current_dir, "test_accounting", "*.xml")):
        c = myCFDI.from_file(f)
        all_invoices[c.uuid] = c

    retenciones = filter_retenciones_iter(all_invoices, ejerc=year)
    retenciones_print(retenciones)


def test_single_html_file():
    all_invoices = []
    for f in glob.iglob(os.path.join(current_dir, "invoices", "*.xml")):
        c = myCFDI.from_file(f)
        all_invoices.append(c)

    res = Representable.html_str_all(all_invoices)
    verify_result(res, "multiple_invoices.html")

    Representable.html_write_all(
        objs=all_invoices,
        target=os.path.join(current_dir, "test_accounting", "multiple_invoices2.html"),
    )

    with open(os.path.join(current_dir, "test_accounting", "multiple_invoices2.html"), "r", encoding='utf-8') as f:
        assert f.read() == res

    # assert filecmp.cmp(
    #     os.path.join(current_dir, "test_accounting", "multiple_invoices.html"),
    #     os.path.join(current_dir, "test_accounting", "multiple_invoices2.html")
    # )



import json
from datetime import datetime
from decimal import Decimal
from unittest import mock

from satcfdi.create.cfd import cfdi40
from satcfdi.pacs import Environment
from satcfdi.pacs.comerciodigital import ComercioDigital
from utils import get_signer, verify_result


def test_comercio_digital_test():
    comercio_digital = ComercioDigital(
        user="AAA010101AAA",
        password="PWD",
        environment=Environment.TEST
    )

    signer = get_signer('xia190128j61')
    emisor = cfdi40.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )

    invoice = cfdi40.Comprobante(
        emisor=emisor,
        lugar_expedicion="27200",
        fecha=datetime.fromisoformat("2022-09-28T22:40:38"),
        receptor=cfdi40.Receptor(
            rfc="ARE100609JP4",
            nombre="ACE RENUEVA SA DE CV",
            uso_cfdi="G03",
            domicilio_fiscal_receptor="85000",
            regimen_fiscal_receptor="601"
        ),
        metodo_pago="PPD",
        forma_pago="99",
        serie="T",
        folio="1000",
        conceptos=cfdi40.Concepto(
            cuenta_predial='1234567890',
            clave_prod_serv='10101702',
            cantidad=Decimal('.10'),
            clave_unidad='E48',
            descripcion='SERVICIOS DE RENTA',
            valor_unitario=Decimal("1.00"),
            impuestos=cfdi40.Impuestos(
                traslados=cfdi40.Impuesto.parse('002|Tasa|0.160000'),
                retenciones=[cfdi40.Impuesto.parse('001|Tasa|0.100000'), cfdi40.Impuesto.parse('002|Tasa|0.106667')],
            ),
            _traslados_incluidos=True
        )
    )
    invoice.sign(signer)

    with mock.patch(f'requests.request') as mk:
        mk.return_value.ok = True
        mk.return_value.headers = {
            'codigo': '000',
            'uuid': '66ea0f5a-4f00-4406-a573-87eb1a34f1fa'
        }
        mk.return_value.content = b'Hello'

        comercio_digital.stamp(
            cfdi=invoice
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'
        mk.call_args.kwargs["data"] = b"".join(mk.call_args.kwargs["data"])

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="test_stamp.json")
        assert verify




from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence

from ..w3.signature import signature_c14n_sha1
from ... import Signer
from ... import CFDI, XElement, ScalarMap


class Folio(ScalarMap):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/1
    Folio fiscal (UUID) a ser cancelado.
    """
    def __init__(
            self,
            uuid: str,
            motivo: str,
            folio_sustitucion: str = None,
    ): 
        """
        Folio fiscal (UUID) a ser cancelado.
        
        :param uuid: UUID generado en la operación de timbrado del comprobante solicitado.
        :param motivo: Clave del motivo de cancelación del comprobante.
        :param folio_sustitucion: Atributo condicional que representa al UUID que sustituye al folio fiscal cancelado. Es requerido cuando la clave del motivo de cancelación es 01.
        """
        
        super().__init__({
            'UUID': uuid,
            'Motivo': motivo,
            'FolioSustitucion': folio_sustitucion,
        })
        

class Cancelacion(XElement):
    """
    Elemento raíz para realizar una solicitud de cancelación de CFDI.
    """
    tag = '{http://www.sat.gob.mx/esquemas/retencionpago/1}Cancelacion'

    def __init__(
            self,
            emisor: Signer,
            folios: Folio | Sequence[Folio | dict],
            fecha: datetime = None
    ):
        """
        Elemento raíz para realizar una solicitud de cancelación de CFDI.

        :param emisor: Atributo requerido para expresar el RFC del emisor del (os) CFDI a cancelar.
        :param fecha: Atributo requerido para expresar la fecha de la operación.
        :param folios: Colección de folios fiscales (UUID) a ser cancelados.
        """
        super().__init__({
            'RfcEmisor': emisor.rfc,
            'Fecha': fecha or datetime.now(),
            'Folios': folios,
        })
        self["_nsmap"] = {
            None: "http://www.sat.gob.mx/esquemas/retencionpago/1",
            "xsd": "http://www.w3.org/2001/XMLSchema",
            "xsi": "http://www.w3.org/2001/XMLSchema-instance"
        }

        sig = signature_c14n_sha1(
            signer=emisor,
            element=self.to_xml(),
            nsmap={
                None: 'http://www.w3.org/2000/09/xmldsig#',
                "xsd": "http://www.w3.org/2001/XMLSchema",
                "xsi": "http://www.w3.org/2001/XMLSchema-instance",
            }
        )
        self['Signature'] = sig




import base64
import logging
import os
from decimal import Decimal
from html import escape as html_escape
from io import BytesIO

import qrcode
from jinja2 import Environment, FileSystemLoader, TemplateNotFound
from jinja2.filters import do_mark_safe
from jinja2.runtime import Undefined
from lxml.etree import QName


# noinspection PyUnresolvedReferences
from .catalog import CATALOGS
from .helpers import desc, format_address, trans, iterate as h_iterate
from ..models import py2html, Code

current_dir = os.path.dirname(__file__)
logger = logging.getLogger(__name__)

complement_order = {
    "TimbreFiscalDigital": 1
}


def finalize(s):
    if isinstance(s, Decimal):
        return s.__format__(",f")

    if isinstance(s, Undefined):
        return do_mark_safe("&nbsp;")

    return s


class PDFEnvironment(Environment):
    @property
    def filter(self):
        def sub(f):
            self.filters[f.__name__] = f
            return f

        return sub

    @property
    def glob(self):
        def sub(f):
            self.globals[f.__name__] = f
            return f

        return sub

    def __init__(self, templates_path=None):
        search_path = [os.path.join(current_dir, 'pdf_templates')]
        if templates_path:
            search_path.insert(0, templates_path)

        super().__init__(
            loader=PDFFIleSystemLoader(searchpath=search_path),
            autoescape=True,
            trim_blocks=True,
            lstrip_blocks=True
        )

        self.filters['desc'] = desc
        self.filters['trans'] = trans
        self.finalize = finalize

        @self.glob
        def iterate(v):
            if isinstance(v, Undefined):
                return v
            return h_iterate(v)

        @self.filter
        def qr_image(url):
            qr = qrcode.QRCode(
                version=1,
                error_correction=qrcode.constants.ERROR_CORRECT_M,
                box_size=5,
                border=0,
            )
            qr.add_data(url)
            qr.make(fit=True)

            img = qr.make_image(fill_color="black", back_color="white")

            with BytesIO() as buffered:
                img.save(buffered, format="PNG")
                return base64.b64encode(buffered.getvalue()).decode("utf-8")

        @self.filter
        def mes(m):
            if isinstance(m, int):
                m = "{:0>2}".format(m)
                return CATALOGS['{http://www.sat.gob.mx/esquemas/retencionpago/1/catalogos}c_Periodo'][m]
            return m

        @self.glob
        def iterate_complements(items):
            def sort_order(x):
                return complement_order.get(QName(x.tag).localname, 0)

            return [(QName(x.tag).localname, x) for x in sorted(iterate(items), key=sort_order)]

        @self.glob
        def tasa_cuota(factor, s):
            if factor == "Tasa":
                return "{:%}".format(s)
            return s

        @self.filter
        def address(k):
            if k:
                address = format_address(k)
                return do_mark_safe(
                    html_escape(address).replace('\n', '<br>')
                )
            return k

        @self.filter
        def dump(data):
            if data:
                res = py2html.dumps(data, escape=True, clubbing=True, translate_keys=trans, default=finalize)
                return do_mark_safe(res)
            return data

        @self.filter
        def simple(data, *args):
            def fields():
                for k in args or data.keys():
                    if v := data.get(k):
                        yield f"<div><b>{html_escape(trans(k))}: </b>{dump(v)}</div>"

            return do_mark_safe(
                "".join(fields())
            )

        @self.filter
        def baa(s):
            r = "&#8203;".join(html_escape(str(s)))
            return do_mark_safe(r)

        @self.filter
        def code(s):
            if isinstance(s, Code):
                return s.code
            return s


class PDFFIleSystemLoader(FileSystemLoader):
    def get_source(
            self, environment: Environment, template: str
    ):
        try:
            return super().get_source(environment, template)
        except TemplateNotFound:
            return "{{ c | dump }}", None, lambda: True


DefaultPDFEnvironment = PDFEnvironment()



import requests
from lxml import etree

from . import PAC, Accept, Document, CancelationAcknowledgment, CancelReason, Environment
from .. import __version__, ResponseError, Signer, XElement
from ..cfdi import CFDI


class ComercioDigital(PAC):
    RFC = "SCD110105654"

    def __init__(self, user: str, password: str, environment=Environment.PRODUCTION):
        super().__init__(environment=environment)
        self.user = user.replace('Ñ', '@')
        self.password = password

    def _request(self, path, data, method="post", headers=None):
        match self.environment:
            case Environment.PRODUCTION:
                host = "https://ws.comercio-digital.mx"
            case Environment.TEST:
                host = "https://pruebas.comercio-digital.mx"
            case _:
                raise NotImplementedError("Not Implemented Environment")

        h = {
            "User-Agent": __version__.__user_agent__,
            "Content-Type": "text/plain",
            "usrws": self.user,
            "pwdws": self.password,
            "tipo": "XML"
        }
        if headers:
            h = h | headers

        r = requests.request(
            method=method,
            url=f"{host}/{path}",
            headers=h,
            data=data,
        )

        if r.ok and 'errmsg' not in r.headers and r.headers['codigo'] == '000':
            return r
        else:
            raise ResponseError(r)

    def issue(self, cfdi: CFDI, accept: Accept = Accept.XML) -> Document:
        raise NotImplementedError()

    def stamp(self, cfdi: CFDI, accept: Accept = Accept.XML) -> Document:
        match cfdi.tag:
            case '{http://www.sat.gob.mx/cfd/3}Comprobante':
                path = "timbre/timbrarV5"
            case '{http://www.sat.gob.mx/cfd/4}Comprobante':
                path = "timbre4/timbrarV5"
            case _:
                raise NotImplementedError("cfdi type not supported")

        if accept & Accept.PDF:
            raise NotImplementedError("accept PDF not supported")

        res = self._request(
            path=path,
            data=iter((
                b'<?xml version="1.0" encoding="utf-8"?>',
                cfdi.xml_bytes(xml_declaration=False)
            ))
        )

        return Document(
            document_id=res.headers['uuid'],
            xml=res.content if accept & Accept.XML else None
        )

    def cancel(self, cfdi: CFDI, reason: CancelReason, substitution_id: str = None, signer: Signer = None) -> CancelationAcknowledgment:
        raise NotImplementedError()

    def recover(self, document_id: str, accept: Accept = Accept.XML) -> Document:
        raise NotImplementedError()

    def rfc_valid(self, rfc: str | list[str]) -> bool | list[bool]:
        raise NotImplementedError()



from collections.abc import Mapping, Sequence
from datetime import date
from itertools import groupby
from typing import Iterable

import xlsxwriter
from tabulate import tabulate

from ._ansi_colors import *
from .formatters import *
from .models import *
from ..create.cfd.catalogos import Impuesto
from ..create.retencion.catalogos import TipoImpuesto

logger = logging.getLogger(__name__)

IVA16 = "002|Tasa|0.160000"
IVA08 = "002|Tasa|0.080000"


def complement_invoices_data(invoices: Mapping[UUID, SatCFDI]):
    for c in invoices.values():
        complement_invoices(invoices, c)


def complement_invoices(invoices: Mapping[UUID, SatCFDI], invoice: SatCFDI):
    c = invoice
    for cfdi_rel in iterate(c.get("CfdiRelacionados")):
        for uuid in cfdi_rel["CfdiRelacionado"]:
            if cfdi := invoices.get(UUID(uuid)):
                cfdi.relations.append(
                    Relation(
                        cfdi_relacionados=cfdi_rel,
                        comprobante=c
                    )
                )

    if c['TipoDeComprobante'] == "P":
        for p in c["Complemento"]["Pagos"]["Pago"]:
            for doc_rel in p.get('DoctoRelacionado', []):
                if cfdi := invoices.get(UUID(doc_rel["IdDocumento"])):
                    cfdi.payments.append(
                        Payment(
                            docto_relacionado=doc_rel,
                            pago=p,
                            comprobante=c
                        )
                    )


def _compare(value, compare_value):
    if compare_value:
        if callable(compare_value):
            return compare_value(value)
        return value == compare_value
    return True


def filter_invoices_iter(
        invoices: Iterable[SatCFDI],
        rfc_emisor=None,
        rfc_receptor=None,
        estatus=None,
        fecha=None,
        invoice_type=None,
        payment_method=None,
        pending_balance=None
):
    for r in invoices:
        if _compare(r["Emisor"]["Rfc"], rfc_emisor) \
                and _compare(r["Receptor"]["Rfc"], rfc_receptor) \
                and _compare(r.estatus, estatus) \
                and _compare(r["Fecha"], fecha) \
                and _compare(r.get("MetodoPago"), payment_method) \
                and _compare(r["TipoDeComprobante"], invoice_type) \
                and _compare(r.saldo_pendiente, pending_balance):
            yield r


def filter_payments_iter(invoices: Mapping[UUID, SatCFDI], rfc_emisor=None, rfc_receptor=None, fecha=None) -> Sequence[PaymentsDetails]:
    for r in filter_invoices_iter(invoices.values(), rfc_emisor=rfc_emisor, rfc_receptor=rfc_receptor, estatus='1', fecha=None):
        match r['TipoDeComprobante']:
            case "I":
                if r['MetodoPago'] == MetodoPago.PAGO_EN_UNA_SOLA_EXHIBICION:
                    if not r.payments:
                        if _compare(r["Fecha"], fecha):
                            yield PaymentsDetails(comprobante=r)
            case "P":
                for p in r["Complemento"]["Pagos"]["Pago"]:
                    if _compare(p['FechaPago'], fecha):
                        for dr in p['DoctoRelacionado']:
                            yield PaymentsDetails(
                                comprobante=r,
                                pago=p,
                                docto_relacionado=dr,
                                comprobante_pagado=invoices[UUID(dr["IdDocumento"])]
                            )


def filter_retenciones_iter(invoices, ejerc: int):
    for a in invoices.values():
        if (a["Version"] == "1.0" and a["Periodo"]["Ejerc"] != ejerc) or \
                (a["Version"] == "2.0" and a["Periodo"]["Ejercicio"] != str(ejerc)):
            continue

        if "Intereses" in a["Complemento"]:
            yield a


def invoice_def():
    # Width, Sum, Header, Value Lambda
    return {
        'Factura': (35, False, format_head),
        'Fecha': (35, False, format_fecha),
        'Emisor': (40, False, format_emisor),
        'Receptor': (40, False, format_receptor),
        'Forma Pago': (40, False, format_forma_pago),

        'SubTotal': (12, True, lambda i: i["SubTotal"]),
        'Descuento': (12, True, lambda i: i.get("Descuento")),
        'IVA16 Tras': (12, True, lambda i: i.get("Impuestos", {}).get("Traslados", {}).get(IVA16, {}).get("Importe")),
        'IVA Ret': (12, True, lambda i: i.get("Impuestos", {}).get("Retenciones", {}).get(Impuesto.IVA, {}).get("Importe")),
        'ISR Ret': (12, True, lambda i: i.get("Impuestos", {}).get("Retenciones", {}).get(Impuesto.ISR, {}).get("Importe")),
        'Total': (12, True, lambda i: i["Total"]),
        'Pendiente': (12, True, lambda i: i.saldo_pendiente or None),
        'Pagos': (35, False, format_pagos),
        'Relaciones': (35, False, format_relaciones),
        'Estado CFDI': (35, False, format_estado_cfdi),
        'Conceptos': (60, False, format_conceptos),
    }


def invoice_confirm_def():
    # Width, Sum, Header, Value Lambda
    return {
        'Factura': (35, False, format_head_pre),
        'Fecha': (35, False, format_fecha),
        'Emisor': (40, False, format_emisor),
        'Receptor': (40, False, format_receptor),
        'Forma Pago': (40, False, format_forma_pago),

        'SubTotal': (12, True, lambda i: i["SubTotal"]),
        'Descuento': (12, True, lambda i: i.get("Descuento")),
        'IVA16 Tras': (12, True, lambda i: i.get("Impuestos", {}).get("Traslados", {}).get(IVA16, {}).get("Importe")),
        'IVA Ret': (12, True, lambda i: i.get("Impuestos", {}).get("Retenciones", {}).get(Impuesto.IVA, {}).get("Importe")),
        'ISR Ret': (12, True, lambda i: i.get("Impuestos", {}).get("Retenciones", {}).get(Impuesto.ISR, {}).get("Importe")),
        'Total': (12, True, lambda i: i["Total"]),
        'Conceptos': (60, False, format_conceptos),
    }


def payment_def():
    # Width, Sum, Header, Value Lambda
    # Lambda 0:payment, 1:factura_pagada
    return {
        'Factura': (35, False, lambda i: format_head(i.comprobante)),
        'Fecha': (35, False, lambda i: format_fecha(i.comprobante)),
        'Emisor': (40, False, lambda i: format_emisor(i.comprobante)),
        'Receptor': (40, False, lambda i: format_receptor(i.comprobante)),
        'Forma Pago': (40, False, lambda i: format_forma_pago_dr(i)),

        'Factura Pagada': (12, False, lambda i: format_head(i.comprobante_pagado) if i.pago else "misma"),
        'Fecha de Pago': (12, False, lambda i: format_fecha_pago(i)),

        'Pagado': (12, False, lambda i: i.total),
        'Saldo Ant': (12, False, lambda i: i.docto_relacionado["ImpSaldoAnt"] if i.pago else i.comprobante["Total"]),
        'Saldo Insoluto': (12, False, lambda i: i.docto_relacionado["ImpSaldoInsoluto"] if i.pago else 0),
        'Parcialidad': (12, False, lambda i: i.docto_relacionado["NumParcialidad"] if i.pago else None),
        'Subtotal': (12, True, lambda i: i.sub_total),
        'Descuento': (12, True, lambda i: i.descuento),
        'IVA16 Tras': (12, True, lambda i: i.impuestos.get("Traslados", {}).get(IVA16, {}).get("Importe")),
        'IVA Ret': (12, True, lambda i: i.impuestos.get("Retenciones", {}).get(Impuesto.IVA, {}).get("Importe")),
        'ISR Ret': (12, True, lambda i: i.impuestos.get("Retenciones", {}).get(Impuesto.ISR, {}).get("Importe")),

        'Estatus SAT': (35, False, lambda i: format_estado_cfdi(i.comprobante)),
    }


def retenciones_def():
    return {
        'RFC de la Institucion': (12, False, lambda i: i["Emisor"].get("RfcE") or i["Emisor"].get("RFCEmisor")),
        'Monto de los intereses nominales': (12, False, lambda i: i["Complemento"]["Intereses"]["MontIntNominal"]),
        'Monto de los intereses reales': (12, False, lambda i: i["Complemento"]["Intereses"]["MontIntReal"]),
        'Perdida': (12, False, lambda i: i["Complemento"]["Intereses"]["Perdida"]),
        'ISR Retenido': (
        12, False, lambda i: sum(x["MontoRet"] for x in i["Totales"]['ImpRetenidos'] if x.get("Impuesto") == TipoImpuesto.ISR) if 'ImpRetenidos' in i["Totales"] else None)
    }


def console_print(invoices, columns):
    headers = ['', *(c for c in columns.keys())]

    def row(n, i):
        return [
            paint("{:0>2}".format(n + 1), COLOR_BACKGROUND_BRIGHT_BLACK if n % 2 == 0 else COLOR_BACKGROUND_BLACK),
            *(f(i) for w, s, f in columns.values())
        ]

    print(
        tabulate(
            [
                row(n, i) for n, i in enumerate(invoices)
            ],
            floatfmt=".2f",
            headers=headers
        )
    )
    print(paint("", COLOR_RESET))


# CONSOLE PRINT
def invoices_confirmation_print(invoices: Sequence[CFDI]):
    console_print(invoices, invoice_confirm_def())


def invoices_print(invoices: Sequence[SatCFDI]):
    console_print(invoices, invoice_def())


def payments_print(payments: Sequence[PaymentsDetails]):
    console_print(payments, payment_def())


def retenciones_print(retentions: Sequence[SatCFDI]):
    console_print(retentions, retenciones_def())


def num2col(n):
    """Number to Excel-style column name, e.g., 0 = A, 25 = Z, 26 = AA, 702 = AAA."""
    n += 1

    name = ''
    while n > 0:
        n, r = divmod(n - 1, 26)
        name = chr(r + ord('A')) + name
    return name


def excel_export(workbook: xlsxwriter.workbook.Workbook, name, invoices, columns):
    worksheet = workbook.add_worksheet(name)

    for n, (w, s, f) in enumerate(columns.values()):
        worksheet.set_column(n, n, w)

    # Title Format
    title_format = workbook.add_format()
    title_format.set_bold()
    title_format.set_pattern(1)
    title_format.set_bg_color('black')
    title_format.set_font_color('white')

    headers = [c for c in columns.keys()]
    for c, cell in enumerate(headers):
        worksheet.write(0, c, cell, title_format)

    text_wrap = workbook.add_format()
    text_wrap.set_text_wrap()
    text_wrap.set_align('top')

    date_format = workbook.add_format({'num_format': 'yyyy-mm-dd'})
    date_format.set_align('top')

    number_format = workbook.add_format({'num_format': '0.00'})
    number_format.set_align('top')

    row_height = 20 * 3
    r = 0
    for r, i in enumerate(invoices):
        worksheet.set_row_pixels(1 + r, row_height)

        for c, cell in enumerate(f(i) for w, s, f in columns.values()):
            if isinstance(cell, Decimal):
                worksheet.write(1 + r, c, cell, number_format)
            elif isinstance(cell, date):
                worksheet.write(1 + r, c, cell, date_format)
            else:
                worksheet.write(1 + r, c, cell, text_wrap)

    title_format.set_num_format('0.00')
    for i, (w, s, f) in enumerate(columns.values()):
        if s:
            col = num2col(i)
            worksheet.write(r + 2, i, '=SUM(' + col + '2:' + col + str(1 + r + 1) + ')', title_format)


# EXPORT TO EXCEL
def invoices_export(workbook, name, invoices: Sequence[SatCFDI]):
    excel_export(workbook, name, invoices, invoice_def())


def payments_export(workbook, name, payments: Sequence[PaymentsDetails]):
    excel_export(workbook, name, payments, payment_def())


def retentions_export(workbook, name, retentions: Sequence[SatCFDI]):
    excel_export(workbook, name, retentions, retenciones_def())


# EXPORT TO TXT
def payments_groupby_receptor(payments: Sequence[PaymentsDetails]):
    res = []
    for receptor, group in groupby(
            sorted(payments, key=lambda r: r.comprobante["Receptor"]["Rfc"]),
            lambda r: r.comprobante["Receptor"]["Rfc"]
    ):
        p = list(group)
        res.append({
            "Receptor": receptor,
            "SubTotal": sum(p.sub_total for p in p),
            "Descuento": sum(p.descuento or 0 for p in p),
            "ISR Ret": sum(p.impuestos.get("Retenciones", {}).get(Impuesto.ISR, {}).get("Importe") or 0 for p in p)
        })
    return res


def payments_retentions_export(file_name, grouped_payments: Sequence):
    with open(file_name, "w", encoding="utf-8") as f:
        def write(line):
            f.write(line)
            f.write("\n")

        write("RFC retenedor|Monto del ingreso recibido|ISR retenido")
        for r in grouped_payments:
            write("{receptor}|{ingreso_recibido}|{isr_retenido}".format(
                receptor=r["Receptor"],
                ingreso_recibido=round(r["SubTotal"] - r["Descuento"], 2),
                isr_retenido=round(r["ISR Ret"], 2)
            ))



import os
from datetime import datetime
from decimal import Decimal

from satcfdi.create.addendas import dvz11
from satcfdi.create.cfd import cfdi40
from satcfdi.pacs.sat import SAT
from tests.utils import get_signer, verify_result, XElementPrettyPrinter

module = 'satcfdi'
current_dir = os.path.dirname(__file__)
current_filename = os.path.splitext(os.path.basename(__file__))[0]
sat = SAT()


def verify_invoice(invoice, path, include_metadata=False):
    if include_metadata:
        verify = verify_result(data=invoice.metadata, filename=f"{path}.txt")
        assert verify

    pp = XElementPrettyPrinter()
    verify = verify_result(data=pp.pformat(invoice), filename=f"{path}.pretty.py")
    assert verify

    verify = verify_result(data=invoice.xml_bytes(pretty_print=True), filename=f"{path}.xml")
    assert verify

    verify = verify_result(data=invoice.html_str(), filename=f"{path}.html")
    assert verify


def test_create_addenda():
    xml_file = "cfdi_addenda"
    signer = get_signer('h&e951128469')
    emisor = cfdi40.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="606",
    )

    invoice = cfdi40.Comprobante(
        emisor=emisor,
        fecha=datetime.fromisoformat("2020-09-29T22:40:38"),
        lugar_expedicion="56820",
        receptor=cfdi40.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03',
            domicilio_fiscal_receptor="59820",
            regimen_fiscal_receptor="601"
        ),
        metodo_pago='PPD',
        serie="A",
        folio="123456",
        conceptos=[
            cfdi40.Concepto(
                cuenta_predial='1234567890',
                clave_prod_serv='10101702',
                cantidad=Decimal('1.00'),
                clave_unidad='E48',
                descripcion='SERVICIOS DE FACTURACION',
                valor_unitario=Decimal('15390.30'),
                impuestos=cfdi40.Impuestos(
                    traslados=cfdi40.Impuesto.parse('002|Tasa|0.160000'),
                    retenciones=[cfdi40.Impuesto.parse('001|Tasa|0.100000'), cfdi40.Impuesto.parse('002|Tasa|0.106667')],
                ),
                _traslados_incluidos=False
            )
        ],
        addenda=dvz11.Diverza(
            generales=dvz11.Generales(
                tipo_documento="Factura"
            )
        )
    )
    invoice.sign(signer=signer)

    verify_invoice(invoice, f"{xml_file}")

    verify_invoice(invoice.process(), f"{xml_file}_process")


def test_copy_cfdi():
    xml_file = "cfdi_copy"
    signer = get_signer('h&e951128469')
    emisor = cfdi40.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="606",
    )

    invoice = cfdi40.Comprobante(
        emisor=emisor,
        fecha=datetime.fromisoformat("2020-09-29T22:40:38"),
        lugar_expedicion="56820",
        receptor=cfdi40.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03',
            domicilio_fiscal_receptor="59820",
            regimen_fiscal_receptor="601"
        ),
        metodo_pago='PPD',
        serie="A",
        folio="123456",
        conceptos=[
            cfdi40.Concepto(
                cuenta_predial='1234567890',
                clave_prod_serv='10101702',
                cantidad=Decimal('1.00'),
                clave_unidad='E48',
                descripcion='SERVICIOS DE FACTURACION',
                valor_unitario=Decimal('15390.30'),
                impuestos=cfdi40.Impuestos(
                    traslados=cfdi40.Impuesto.parse('002|Tasa|0.160000'),
                    retenciones=[cfdi40.Impuesto.parse('001|Tasa|0.100000'), cfdi40.Impuesto.parse('002|Tasa|0.106667')],
                ),
                _traslados_incluidos=False
            )
        ]
    )
    invoice.sign(signer=signer)

    copy = invoice.copy()
    verify_invoice(copy, f"{xml_file}")

    copy.json_str(pretty_print=True)



class CFDIError(Exception):
    """
    CFDI Error
    """


class SchemaValidationError(Exception):
    """
    Schema Validation Error
    """

    def __init__(self, error_log):
        self.error_log = error_log
        super().__init__(error_log)


class ResponseError(Exception):
    """
    Request Response Error
    """

    def __init__(self, response):
        self.response = response
        super().__init__(response)


class NamespaceMismatchError(Exception):
    """
    Returned by objectify and xmlify
    """

    def __init__(self, node):
        self.node = node
        super().__init__(node)


class DocumentNotFoundError(ResponseError):
    pass


class CFDIInvalidError(ResponseError):
    pass



TRANSFORMS = {
    '{www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarCtas}AuxiliarCtas': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/AuxiliarCtas/AuxiliarCtas_{v}.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarCtas}AuxiliarCtas': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/AuxiliarCtas/AuxiliarCtas_{v}.xslt',
    '{www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarFolios}RepAuxFol': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/AuxiliarFolios/AuxiliarFolios_{v}.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarFolios}RepAuxFol': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/AuxiliarFolios/AuxiliarFolios_{v}.xslt',
    '{www.sat.gob.mx/esquemas/ContabilidadE/1_1/BalanzaComprobacion}Balanza': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/BalanzaComprobacion/BalanzaComprobacion_{v}.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_1/BalanzaComprobacion}Balanza': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/BalanzaComprobacion/BalanzaComprobacion_{v}.xslt',
    '{www.sat.gob.mx/esquemas/ContabilidadE/1_1/CatalogoCuentas}Catalogo': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/CatalogoCuentas/CatalogoCuentas_{v}.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_1/CatalogoCuentas}Catalogo': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/CatalogoCuentas/CatalogoCuentas_{v}.xslt',
    '{www.sat.gob.mx/esquemas/ContabilidadE/1_1/PolizasPeriodo}Polizas': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/PolizasPeriodo/PolizasPeriodo_{v}.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_1/PolizasPeriodo}Polizas': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/PolizasPeriodo/PolizasPeriodo_{v}.xslt',
    '{www.sat.gob.mx/esquemas/ContabilidadE/1_1/SelloDigitalContElec}SelloDigitalContElec': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/SelloDigitalContElec/SelloDigitalContElec.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_1/SelloDigitalContElec}SelloDigitalContElec': 'www.sat.gob.mx/esquemas/ContabilidadE/{v}/SelloDigitalContElec/SelloDigitalContElec.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_3/AuxiliarCtas}AuxiliarCtas': 'www.sat.gob.mx/esquemas/ContabilidadE/1_3/AuxiliarCtas/AuxiliarCtas_1_2.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_3/AuxiliarFolios}RepAuxFol': 'www.sat.gob.mx/esquemas/ContabilidadE/1_3/AuxiliarFolios/AuxiliarFolios_1_2.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_3/BalanzaComprobacion}Balanza': 'www.sat.gob.mx/esquemas/ContabilidadE/1_3/BalanzaComprobacion/BalanzaComprobacion_1_2.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_3/CatalogoCuentas}Catalogo': 'www.sat.gob.mx/esquemas/ContabilidadE/1_3/CatalogoCuentas/CatalogoCuentas_1_2.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_3/PolizasPeriodo}Polizas': 'www.sat.gob.mx/esquemas/ContabilidadE/1_3/PolizasPeriodo/PolizasPeriodo_1_2.xslt',
    '{www.sat.gob.mx/esquemas/ContabilidadE/1_3/SelloDigitalContElec}SelloDigitalContElec': 'www.sat.gob.mx/esquemas/ContabilidadE/1_3/SelloDigitalContElec/SelloDigitalContElec.xslt',
    '{http://www.sat.gob.mx/esquemas/ContabilidadE/1_3/SelloDigitalContElec}SelloDigitalContElec': 'www.sat.gob.mx/esquemas/ContabilidadE/1_3/SelloDigitalContElec/SelloDigitalContElec.xslt',
    '{http://www.sat.gob.mx/esquemas/retencionpago/1}Retenciones': 'www.sat.gob.mx/esquemas/retencionpago/1/retenciones.xslt',
    '{http://www.sat.gob.mx/esquemas/retencionpago/2}Retenciones': 'www.sat.gob.mx/esquemas/retencionpago/2/retenciones.xslt',
    '{http://www.sat.gob.mx/cfd/3}Comprobante': 'www.sat.gob.mx/sitio_internet/cfd/3/cadenaoriginal_{v}/cadenaoriginal_{v}.xslt',
    '{http://www.sat.gob.mx/cfd/4}Comprobante': 'www.sat.gob.mx/sitio_internet/cfd/4/cadenaoriginal_{v}/cadenaoriginal_{v}.xslt',
    '{http://www.sat.gob.mx/TimbreFiscalDigital}TimbreFiscalDigital': 'www.sat.gob.mx/sitio_internet/cfd/TimbreFiscalDigital/cadenaoriginal_TFD_{v}.xslt',
    '{http://www.sat.gob.mx/vehiculousado}VehiculoUsado': 'www.sat.gob.mx/sitio_internet/cfd/vehiculousado/vehiculousado.xslt',
}



import json
import pickle
from time import time

import requests
import urllib3
from requests.structures import CaseInsensitiveDict

from .utils import get_form, generate_token, request_ref_headers, request_verification_token, random_ajax_id
from .. import Signer, ResponseError


class PortalManager(requests.Session):
    def __init__(self, signer: Signer):
        super().__init__()
        try:
            urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH'
        except:
            pass
        self.signer = signer

        self.headers = CaseInsensitiveDict(
            {
                "User-Agent": 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36',
                "Accept-Encoding": 'gzip, deflate, br',
                "Accept": 'text/html,application/xhtml+xml,application/xml',
                "Connection": "keep-alive",
                'Pragma': 'no-cache',
                'Cache-Control': 'no-cache',
            }
        )

    def save_session(self, target):
        pickle.dump(self.cookies, target)

    def load_session(self, source):
        self.cookies.update(pickle.load(source))

    def form(self, action, referer_url, data):
        res = self.post(
            url=action,
            headers=request_ref_headers(referer_url),
            data=data
        )
        assert res.status_code == 200
        return res

    def fiel_login(self, login_response):
        action, data = get_form(login_response, id='certform')
        return self.form(
            action,
            login_response.request.url,
            data | {
                'token': generate_token(self.signer, code=data['guid']),
                'fert': self.signer.certificate.get_notAfter()[2:].decode(),
            }
        )


class SATPortal(PortalManager):
    BASE_URL = 'https://loginda.siat.sat.gob.mx'

    def login(self):
        res = self.get(
            url=f'{self.BASE_URL}/nidp/app/login?id=fiel'
        )
        assert res.status_code == 200
        action, data = get_form(res)

        return self.fiel_login(
            login_response=self.form(action, res.request.url, data)
        )

    def home_page(self):
        return self.get(
            url=f'{self.BASE_URL}/nidp/app?sid=0'
        )

    def logout(self):
        return self.get(
            url=f'{self.BASE_URL}/nidp/app/logout',
            headers={
                'referer': f'{self.BASE_URL}/nidp/app?sid=0'
            },
            allow_redirects=False
        )

    def declaraciones_provisionales_login(self):
        res = self.get(
            url='https://ptscdecprov.clouda.sat.gob.mx',
            allow_redirects=True
        )
        assert res.status_code == 200

        action, data = get_form(res)
        res = self.form(action, res.request.url, data)
        return res


class SATFacturaElectronica(PortalManager):
    BASE_URL = 'https://portal.facturaelectronica.sat.gob.mx'
    REQUEST_CONTEXT = 'appId=cid-v1:20ff76f4-0bca-495f-b7fd-09ca520e39f7'

    def __init__(self, signer: Signer):
        super().__init__(signer)
        self._ajax_id = random_ajax_id()
        self._request_verification_token = None

    def login(self):
        res = self.get(
            url=self.BASE_URL
        )
        assert res.status_code == 200
        try:
            action, data = get_form(res)
        except IndexError as ex:
            raise ValueError("Login form not found, please try again") from ex

        if action.startswith('https://cfdiau.sat.gob.mx/'):
            assert 'nidp/wsfed/ep?id=SATUPCFDiCon' in action

            res = self.fiel_login(
                login_response=self.form(
                    action.replace('nidp/wsfed/ep?id=SATUPCFDiCon', 'nidp/app/login?id=SATx509Custom'),
                    res.request.url,
                    data
                )
            )

            action, data = get_form(res)
            res = self.form(action, res.request.url, data)

            action, data = get_form(res)

        res = self.form(action, res.request.url, data)

        self._request_verification_token = request_verification_token(res)
        self._ajax_id = random_ajax_id()
        return res

    def _reload_verification_token(self):
        res = self.get(
            url=f'{self.BASE_URL}/Factura/GeneraFactura',
            allow_redirects=False
        )
        if res.status_code == 200:
            self._request_verification_token = request_verification_token(res)
        else:
            raise ValueError('Please Login Again')

    def reactivate_session(self):
        res = self.post(
            url=f'{self.BASE_URL}/Home/ReActiveSession',
            headers={
                'Origin': self.BASE_URL,
                'Request-Context': self.REQUEST_CONTEXT,
                'Request-Id': f'|{self._ajax_id}.{random_ajax_id()}'
            },
            allow_redirects=False
        )
        return res

    def _request(self, method, path, data=None, params=None):
        if self._request_verification_token is None:
            self._reload_verification_token()

        res = self.request(
            method=method,
            url=f'{self.BASE_URL}/{path}',
            headers={
                'Origin': self.BASE_URL,
                'Authority': self.BASE_URL,
                'Request-Context': self.REQUEST_CONTEXT,
                '__RequestVerificationToken': self._request_verification_token,
                'Request-Id': f'|{self._ajax_id}.{random_ajax_id()}'  # |pR4Px.o0yAS
            },
            data=data,
            params=params,
            allow_redirects=False
        )
        if res.status_code == 200:
            return res.json()
        else:
            raise ResponseError(res)

    def legal_name_valid(self, rfc, legal_name):
        res = self._request(
            method='POST',
            path='Clientes/ValidaRazonSocialRFC',
            data={
                'rfcValidar': rfc.upper(),
                'razonSocial': legal_name.upper(),
            })
        if not res['exitoso']:
            raise ResponseError(res)
        return res['resultado']

    def rfc_valid(self, rfc):
        res = self._request(
            method='POST',
            path='Clientes/ExisteLrfc',
            data={
                'rfcValidar': rfc.upper()
            }
        )
        if not res['exitoso']:
            raise ResponseError(res)
        return res['resultado']

    def lco_details(self, rfc, apply_border_region=True):
        res = self._request(
            method='GET',
            path='Clientes/ValidaLco',
            params={
                'rfcValidar': rfc.upper(),
                'aplicaRegionFronteriza': apply_border_region,
                "_": int(time() * 1000)
            }
        )
        return json.loads(res)



import logging
import os
import lxml.etree
from weasyprint import CSS

from . import CFDIError, XElement
from .transform import *
from .transform.helpers import fmt_decimal

current_dir = os.path.dirname(__file__)
logger = logging.getLogger(__name__)

VERIFICA_CFDI = "https://verificacfdi.facturaelectronica.sat.gob.mx/default.aspx"
VERIFICA_RETENCION = "https://prodretencionverificacion.clouda.sat.gob.mx/"
PDF_CSS = CSS(string="@page {margin: 1.0cm 1.27cm 1.1cm 0.85cm;}")


class CFDI(XElement):
    def to_xml(self, validate=False, include_schema_location=True) -> lxml.etree.Element:
        return super().to_xml(validate, include_schema_location)

    def xml_bytes(self, pretty_print=False, xml_declaration=True, validate=False, include_schema_location=True) -> bytes:
        return super().xml_bytes(pretty_print=pretty_print, xml_declaration=xml_declaration, validate=validate, include_schema_location=include_schema_location)

    def process(self, validate=False) -> 'CFDI':
        return CFDI.from_xml(self.to_xml(validate=validate))

    def copy(self) -> 'CFDI':
        cfdi = CFDI(super().copy())
        cfdi.tag = self.tag
        return cfdi

    @property
    def verifica_url(self) -> str:
        match self.tag:
            case '{http://www.sat.gob.mx/cfd/3}Comprobante' | '{http://www.sat.gob.mx/cfd/4}Comprobante':
                base_url = VERIFICA_CFDI
                q = {
                    "id": self['Complemento']['TimbreFiscalDigital']["UUID"],
                    "re": self['Emisor']["Rfc"],
                    "rr": self['Receptor']["Rfc"],
                    "tt": fmt_decimal(self["Total"]),
                    "fe": self["Sello"][-8:]
                }

            case '{http://www.sat.gob.mx/esquemas/retencionpago/1}Retenciones':
                base_url = VERIFICA_RETENCION
                q = {
                    "id": self["Complemento"]["TimbreFiscalDigital"]["UUID"],
                    "re": self['Emisor']["RFCEmisor"],
                    "tt": fmt_decimal(self['Totales']["MontoTotOperacion"]),
                    "fe": self["Sello"][-8:]
                }
                match self['Receptor']['Nacionalidad']:
                    case "Nacional":
                        q["rr"] = self['Receptor']['Nacional']["RFCRecep"]
                    case "Extranjero":
                        if nr := self['Receptor']['Extranjero'].get("NumRegIdTrib"):
                            q["nr"] = nr

            case '{http://www.sat.gob.mx/esquemas/retencionpago/2}Retenciones':
                base_url = VERIFICA_RETENCION
                q = {
                    "id": self["Complemento"]["TimbreFiscalDigital"]["UUID"],
                    "re": self['Emisor']["RfcE"],
                    "tt": fmt_decimal(self['Totales']["MontoTotOperacion"]),
                    "fe": self["Sello"][-8:]
                }
                match self['Receptor']['NacionalidadR']:
                    case "Nacional":
                        q["rr"] = self['Receptor']['Nacional']["RfcR"]
                    case "Extranjero":
                        if nr := self['Receptor']['Extranjero'].get("NumRegIdTribR"):
                            q["nr"] = nr

            case _:
                raise CFDIError("No Verification URL found")

        if base_url:
            return base_url + "?" + "&".join(f"{k}={v}" for k, v in q.items())

    def cadena_original(self):
        transform = xslt_transform(self.tag, self['Version'])
        xml = super().to_xml()
        return str(transform(xml))



import enum
from collections.abc import Sequence, Mapping

from lxml import etree

parser = etree.XMLParser(no_network=True, remove_comments=True, remove_blank_text=True, huge_tree=True, collect_ids=False, remove_pis=True)


class ScalarMap(dict):
    pass


def iterate(item):
    if isinstance(item, str | bytes | ScalarMap):
        return [item]
    if isinstance(item, Mapping):
        return item.values()
    if isinstance(item, Sequence):
        return item
    if item is None:
        return []
    return [item]


class StrEnum(str, enum.Enum):  # Compatible with Python 3.10
    def __str__(self):
        return self.value

    def __format__(self, format_spec):
        return self.value.__format__(format_spec)

    @classmethod
    def get(cls, key, default=None):
        return cls._member_map_.get(key, default)



"""terceros http://www.sat.gob.mx/terceros"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class TInformacionAduanera(ScalarMap):
    """
    Tipo definido para expresar información aduanera
    
    :param numero: Atributo requerido para expresar el número del documento aduanero que ampara la importación del bien.
    :param fecha: Atributo requerido para expresar la fecha de expedición del documento aduanero que ampara la importación del bien.
    :param aduana: Atributo opcional para precisar la aduana por la que se efectuó la importación del bien.
    """
    
    def __init__(
            self,
            numero: str,
            fecha: date,
            aduana: str = None,
    ): 
        super().__init__({
            'Numero': numero,
            'Fecha': fecha,
            'Aduana': aduana,
        })
        

class TUbicacionFiscal(ScalarMap):
    """
    Tipo definido para expresar domicilios o direcciones
    
    :param calle: Este atributo requerido sirve para precisar la avenida, calle, camino o carretera donde se da la ubicación.
    :param municipio: Atributo requerido que sirve para precisar el municipio o delegación (en el caso del Distrito Federal) en donde se da la ubicación.
    :param estado: Atributo requerido que sirve para precisar el estado o entidad federativa donde se da la ubicación.
    :param pais: Atributo requerido que sirve para precisar el país donde se da la ubicación.
    :param codigo_postal: Atributo requerido que sirve para asentar el código postal en donde se da la ubicación.
    :param no_exterior: Este atributo opcional sirve para expresar el número particular en donde se da la ubicación sobre una calle dada.
    :param no_interior: Este atributo opcional sirve para expresar información adicional para especificar la ubicación cuando calle y número exterior (noExterior) no resulten suficientes para determinar la ubicación de forma precisa.
    :param colonia: Este atributo opcional sirve para precisar la colonia en donde se da la ubicación cuando se desea ser más específico en casos de ubicaciones urbanas.
    :param localidad: Atributo opcional que sirve para precisar la ciudad o población donde se da la ubicación.
    :param referencia: Atributo opcional para expresar una referencia de ubicación adicional.
    """
    
    def __init__(
            self,
            calle: str,
            municipio: str,
            estado: str,
            pais: str,
            codigo_postal: str,
            no_exterior: str = None,
            no_interior: str = None,
            colonia: str = None,
            localidad: str = None,
            referencia: str = None,
    ): 
        super().__init__({
            'Calle': calle,
            'Municipio': municipio,
            'Estado': estado,
            'Pais': pais,
            'CodigoPostal': codigo_postal,
            'NoExterior': no_exterior,
            'NoInterior': no_interior,
            'Colonia': colonia,
            'Localidad': localidad,
            'Referencia': referencia,
        })
        

class Parte(ScalarMap):
    """
    Nodo opcional para expresar las partes o componentes que integran la totalidad del concepto expresado en el CFD o CFDI
    
    :param cantidad: Atributo requerido para precisar la cantidad de bienes o servicios del tipo particular definido por la presente parte.
    :param descripcion: Atributo requerido para precisar la descripción del bien o servicio cubierto por la presente parte.
    :param unidad: Atributo opcional para precisar la unidad de medida aplicable para la cantidad expresada en la parte.
    :param no_identificacion: Atributo opcional para expresar el número de serie del bien o identificador del servicio amparado por la presente parte.
    :param valor_unitario: Atributo opcional para precisar el valor o precio unitario del bien o servicio cubierto por la presente parte.
    :param importe: Atributo opcional para precisar el importe total de los bienes o servicios de la presente parte. Debe ser equivalente al resultado de multiplicar la cantidad por el valor unitario expresado en la parte.
    :param informacion_aduanera: Nodo opcional para introducir la información aduanera aplicable cuando se trate de partes o componentes importados vendidos de primera mano.
    """
    
    def __init__(
            self,
            cantidad: Decimal | int,
            descripcion: str,
            unidad: str = None,
            no_identificacion: str = None,
            valor_unitario: Decimal | int = None,
            importe: Decimal | int = None,
            informacion_aduanera: TInformacionAduanera | dict | Sequence[TInformacionAduanera | dict] = None,
    ): 
        super().__init__({
            'Cantidad': cantidad,
            'Descripcion': descripcion,
            'Unidad': unidad,
            'NoIdentificacion': no_identificacion,
            'ValorUnitario': valor_unitario,
            'Importe': importe,
            'InformacionAduanera': informacion_aduanera,
        })
        

class InformacionAduanera(ScalarMap):
    """
    Nodo opcional para introducir la información aduanera aplicable cuando se trate de ventas de primera mano de mercancías importadas.
    
    """
    
    def __init__(
            self,
            hello_there_super: TInformacionAduanera | dict
    ): 
        super().__init__({
        })
        

class InformacionFiscalTercero(ScalarMap):
    """
    Nodo opcional para expresar información fiscal de terceros
    
    """
    
    def __init__(
            self,
            hello_there_super: TUbicacionFiscal | dict
    ): 
        super().__init__({
        })
        

class Traslado(ScalarMap):
    """
    Nodo para la información detallada de un traslado de impuesto específico
    
    :param impuesto: Atributo requerido para señalar el tipo de impuesto trasladado
    :param tasa: Atributo requerido para señalar la tasa del impuesto que se traslada por cada concepto amparado en el comprobante
    :param importe: Atributo requerido para señalar el importe del impuesto trasladado
    """
    
    def __init__(
            self,
            impuesto: str,
            tasa: Decimal | int,
            importe: Decimal | int,
    ): 
        super().__init__({
            'Impuesto': impuesto,
            'Tasa': tasa,
            'Importe': importe,
        })
        

class Retencion(ScalarMap):
    """
    Nodo para la información detallada de una retención de impuesto específico
    
    :param impuesto: Atributo requerido para señalar el tipo de impuesto retenido
    :param importe: Atributo requerido para señalar el importe o monto del impuesto retenido
    """
    
    def __init__(
            self,
            impuesto: str,
            importe: Decimal | int,
    ): 
        super().__init__({
            'Impuesto': impuesto,
            'Importe': importe,
        })
        

class Impuestos(ScalarMap):
    """
    Nodo requerido para capturar los impuestos aplicables.
    
    :param retenciones: Nodo opcional para capturar los impuestos retenidos aplicables
    :param traslados: Nodo opcional para asentar o referir los impuestos trasladados aplicables
    """
    
    def __init__(
            self,
            retenciones: Retencion | dict | Sequence[Retencion | dict] = None,
            traslados: Traslado | dict | Sequence[Traslado | dict] = None,
    ): 
        super().__init__({
            'Retenciones': retenciones,
            'Traslados': traslados,
        })
        

class PorCuentadeTerceros(CFDI):
    """
    Complemento concepto para la emisión de Comprobante Fiscal Digital (CFD) y Comprobante Fiscal Digital a través de Internet (CFDI) por orden y cuenta de terceros.
    
    :param rfc: Atributo requerido para la Clave del Registro Federal de Contribuyentes correspondiente al contribuyente emisor del comprobante sin guiones o espacios.
    :param impuestos: Nodo requerido para capturar los impuestos aplicables.
    :param nombre: Atributo opcional para el nombre o razón social del contribuyente emisor del comprobante.
    :param informacion_fiscal_tercero: Nodo opcional para expresar información fiscal de terceros
    :param informacion_aduanera: Nodo opcional para introducir la información aduanera aplicable cuando se trate de ventas de primera mano de mercancías importadas.
    :param parte: Nodo opcional para expresar las partes o componentes que integran la totalidad del concepto expresado en el CFD o CFDI
    :param cuenta_predial: Nodo opcional para asentar el número de cuenta predial con el que fue registrado el inmueble, en el sistema catastral de la entidad federativa de que trate.
    """
    
    tag = '{http://www.sat.gob.mx/terceros}PorCuentadeTerceros'
    version = '1.1'
    
    def __init__(
            self,
            rfc: str,
            impuestos: Impuestos | dict,
            nombre: str = None,
            informacion_fiscal_tercero: InformacionFiscalTercero | dict = None,
            informacion_aduanera: InformacionAduanera | dict = None,
            parte: Parte | dict | Sequence[Parte | dict] = None,
            cuenta_predial: str = None,
    ): 
        super().__init__({
            'Version': self.version,
            'Rfc': rfc,
            'Impuestos': impuestos,
            'Nombre': nombre,
            'InformacionFiscalTercero': informacion_fiscal_tercero,
            'InformacionAduanera': informacion_aduanera,
            'Parte': parte,
            'CuentaPredial': cuenta_predial,
        })
        




import json
from lxml import etree

from .printer import Representable
from .transform import SchemaCollector, cfdi_schemas, validate_xsd
from .utils import ScalarMap, parser
from .transform.objectify import cfdi_objectify
from .transform.xmlify import cfdi_xmlify


class XElement(ScalarMap, Representable):
    tag = None

    def to_xml(self, validate=False, include_schema_location=False) -> etree.Element:
        xml = cfdi_xmlify[self.tag](self)

        if validate or include_schema_location:
            col = SchemaCollector()
            cfdi_schemas[self.tag](col, self)
            if validate:
                validate_xsd(xml, col.base)
            if include_schema_location:
                xml.attrib['{http://www.w3.org/2001/XMLSchema-instance}schemaLocation'] = " ".join(col.schemas)

        return xml

    def process(self, validate=False) -> 'XElement':
        return XElement.from_xml(self.to_xml(validate=validate))

    def copy(self) -> 'XElement':
        el = XElement(super().copy())
        el.tag = self.tag
        return el

    @classmethod
    def from_xml(cls, xml_root) -> 'XElement':
        obj = cfdi_objectify[xml_root.tag](cls, xml_root)
        if not isinstance(obj, XElement):
            obj = XElement(obj)
            obj.tag = xml_root.tag
        return obj

    @classmethod
    def from_file(cls, filename) -> 'XElement':
        return cls.from_xml(etree.parse(filename, parser=parser).getroot())

    @classmethod
    def from_string(cls, string) -> 'XElement':
        return cls.from_xml(etree.fromstring(string, parser=parser))

    def xml_write(self, target, pretty_print=False, xml_declaration=True):
        xml = self.to_xml()
        et = etree.ElementTree(xml)
        et.write(
            target,
            xml_declaration=xml_declaration,
            encoding="UTF-8",
            pretty_print=pretty_print
        )

    def xml_bytes(self, pretty_print=False, xml_declaration=True, validate=False, include_schema_location=False) -> bytes:
        xml = self.to_xml(validate=validate, include_schema_location=include_schema_location)
        return etree.tostring(xml, xml_declaration=xml_declaration, encoding="UTF-8", pretty_print=pretty_print)

    def json_write(self, target, pretty_print=False):
        if isinstance(target, str):
            with open(target, 'w') as f:
                json.dump(self, f, ensure_ascii=False, default=str, indent=2 if pretty_print else None)
                return

        json.dump(self, target, ensure_ascii=False, default=str, indent=2 if pretty_print else None)

    def json_str(self, pretty_print=False) -> str:
        return json.dumps(self, ensure_ascii=False, default=str, indent=2 if pretty_print else None)

    def __repr__(self):
        # return '%s.%s(%s)' % (self.__class__.__module__,
        #                       self.__class__.__qualname__,
        #                       f'{repr(self.tag)}, {super().__repr__()}')
        return '%s(%s)' % (
            self.__class__.__qualname__,
            f'{dict.__repr__(self)}'
        )



from dataclasses import dataclass
from enum import Flag, auto, Enum
from .. import CFDI, Signer
from ..create.cancela import cancelacion
from ..create.cancela import cancelacionretencion
from ..create.cancela.aceptacionrechazo import SolicitudAceptacionRechazo

__all__ = [
    'PAC',
    'Accept',
    'Document',
    'CancelationAcknowledgment',
    'AcceptRejectAcknowledgment',
    'CancelReason',
    'TaxpayerStatus',
    'Environment'
]


class Accept(Flag):
    NOTHING = 0
    XML = auto()
    PDF = auto()
    XML_PDF = XML | PDF


@dataclass(init=True)
class Document:
    document_id: str
    xml: bytes = None
    pdf: bytes = None

    @property
    def cfdi(self) -> CFDI:
        return CFDI.from_string(self.xml)


@dataclass(init=True)
class CancelationAcknowledgment:
    code: str | dict
    acuse: bytes = None


@dataclass(init=True)
class AcceptRejectAcknowledgment:
    folios: dict
    acuse: bytes = None


def _enum_value(val):
    if isinstance(val, Enum):
        return val.value
    return val


class CancelReason(Enum):
    COMPROBANTE_EMITIDO_CON_ERRORES_CON_RELACION = "01"
    """
    -- Comprobante emitido con errores con relación --
    
    Este supuesto aplica cuando la factura generada contiene un error en
    la clave del producto, valor unitario, descuento o cualquier otro dato, por lo que se
    debe reexpedir. En este caso, primero se sustituye la factura y cuando se solicita la
    cancelación, se incorpora el folio de la factura que sustituye a la cancelada.

    ¿Qué debo hacer si uso el motivo de cancelación “01” Comprobantes emitidos con
    errores con relación y el comprobante no se cancela o presenta error?

    Se podrá utilizar la clave 02 para realizar la cancelación de los CFDI
    relacionados incluyendo el que sustituye al CFDI a cancelar, esto con la finalidad de
    que no se genere un estatus de “No cancelable”

    En los casos en los que subsista la operación, se deberá emitir un nuevo comprobante
    con la información correcta y la clave de tipo relación 04 sustitución de CFDI previos
    relacionando el folio fiscal del comprobante que se sustituye    
    """

    COMPROBANTE_EMITIDO_CON_ERRORES_SIN_RELACION = "02"
    """
    -- Comprobante emitido con errores sin relación --
    
    Este supuesto aplica cuando la factura generada contiene un error en
    la clave del producto, valor unitario, descuento o cualquier otro dato y no se requiera
    relacionar con otra factura generada.
    """

    NO_SE_LLEVO_A_CABO_LA_OPERACION = "03"
    """
    -- No se llevó a cabo la operación --
    
    Este supuesto aplica cuando se facturó una operación que no se
    concreta.
    """
    OPERACION_NORMATIVA_RELACIONADA_EN_LA_FACTURA_GLOBAL = "04"
    """
    -- Operación nominativa relacionada en la factura global --
    
    Este supuesto aplica cuando se incluye una venta en la factura global
    de operaciones con el público en general y posterior a ello, el cliente solicita su factura
    nominativa, lo que conlleva a cancelar la factura global y reexpedirla, así como
    generar la factura nominativa al cliente.
    """


class TaxpayerStatus(Enum):
    PRESUNTO = 'Presunto'
    """
    PRESUNTO: Cuando la autoridad fiscal detecta que un contribuyente ha estado emitiendo comprobantes de manera irregular,
    se presumirá la inexistencia de las operaciones amparadas en tales comprobantes.

    En este supuesto, procederá a notificar a los contribuyentes que se encuentren en dicha situación a través de su buzón
    tributario, de la página de internet del Servicio de Administración Tributaria, así como mediante una publicación en el
    Diario Oficial de la Federación.
    """

    SENTENCIA_FAVORABLE = 'Sentencia Favorable'
    """
    SENTENCIA FAVORABLE: El contribuyente tiene treinta días para acreditar que efectivamente sus operaciones son reales, si 
    logra corregir su situación se publica este estado.
    """

    DEFINITIVO = 'Definitivo'
    """
    DEFINITIVO: Si el contribuyente no atiende el llamado de la autoridad en quince días a partir de la última de las
    notificaciones o el contribuyente no pueda desvirtuar la existencia de sus operaciones se publica este estado.
    """

    DESVIRTUADO = 'Desvirtuado'
    """
    DESVIRTUADO: Cuando el contribuyente aporte a la autoridad fiscal la documentación e información que consideren 
    pertinentes para desvirtuar los hechos que llevaron a notificarlos con un plazo de quince días contados a partir de la
    última de las notificaciones que se hayan efectuado.
    """


class Environment(Enum):
    PRODUCTION = auto()
    TEST = auto()


class PAC:
    RFC = None

    def __init__(self, environment):
        self.environment = environment

    def status(self, cfdi: CFDI) -> dict:
        """
        Consulta el estado de un CFDI
        :return: Respuesta de la consulta
        """
        raise NotImplementedError()

    def validate(self, cfdi: CFDI):
        raise NotImplementedError()

    def issue(self, cfdi: CFDI, accept: Accept = Accept.XML) -> Document:
        """
        Operation to request CFDI be sealed and stamped by PAC
        :param accept:
        :param cfdi:
        :return:
        """
        raise NotImplementedError()

    def stamp(self, cfdi: CFDI, accept: Accept = Accept.XML) -> Document:
        """
        Operation to request sealed CFDI be stamped by PAC
        :param accept:
        :param cfdi:
        :return:
        document_id and bytes of stamped xml
        """
        raise NotImplementedError()

    def recover(self, document_id: str, accept: Accept = Accept.XML) -> Document:
        """

        :param accept:
        :param document_id:
        :return:
        """
        raise NotImplementedError()

    def cancel(self, cfdi: CFDI, reason: CancelReason, substitution_id: str = None, signer: Signer = None) -> CancelationAcknowledgment:
        """
        Operation to request single cfdi to be canceled
        :param signer:
        :param cfdi:
        :param substitution_id:
        :param reason:
        :return:
        """
        raise NotImplementedError()

    def cancel_comprobante(self, cancelation: cancelacion.Cancelacion) -> CancelationAcknowledgment:
        """
        Operation to Cancel a Comprobante
        """
        raise NotImplementedError()

    def cancel_retencion(self, cancelation: cancelacionretencion.Cancelacion) -> CancelationAcknowledgment:
        """
        Operation to Cancel a Retencion
        """
        raise NotImplementedError()

    def accept_reject(self, request: SolicitudAceptacionRechazo) -> AcceptRejectAcknowledgment:
        """
        Operation to Accept Reject a Cancellation Request
        """
        raise NotImplementedError()

    def pending(self, rfc: str) -> list[str]:
        """
        Operation to get pending cancellations
        """
        raise NotImplementedError()

    def list_69b(self, rfc: str) -> TaxpayerStatus | None:
        """
        Operation to get list69b status
        """
        raise NotImplementedError()



import logging
import os
from datetime import date, datetime
from itertools import chain
from unittest import mock

import pytest
from OpenSSL import crypto

from satcfdi import Code
from satcfdi.models.certificate import CertificateType
from satcfdi.models.certificate_request import CertificateSigningRequest
from satcfdi.exceptions import CFDIError
from satcfdi.models.curp import CURP
from satcfdi.models.date_period import DatePeriod
from satcfdi.models.rfc import RFC, RFCType, RFC_Generico_Nacional, RFC_Generico_Extranjero
from satcfdi.models.signer import Signer
from satcfdi.pacs.sat import SAT
from satcfdi.transform import verify_certificate
from satcfdi.utils import StrEnum
from tests.constants import PERSONAS_FISICAS, PERSONAS_MORALES
from tests.utils import get_signer, SAT_Certificate_Store_Pruebas

module = 'satcfdi'
current_dir = os.path.dirname(__file__)

logging.basicConfig(level=logging.INFO)

sat = SAT()


def test_descarga_certificado_sat_pac():
    cert = sat.recover_certificate("00001000000504465028")

    # assert cert.issuer().startswith("1.2.840.113549.1.9.2=responsable: ADMINISTRACION CENTRAL DE SERVICIOS TRIBUTARIOS AL CONTRIBUYE")
    # assert cert.subject().startswith(r"OU=MEGAPACSAT970701NN301,2.5.4.5=\ / GARJ750416HDFRVR09,2.5.4.45=SAT970701NN3 / GARJ7504168N4,O=SERVIC")
    # assert cert.issuer().startswith("CN=AUTORIDAD CERTIFICADORA,O=SERVICIO DE ADMINISTRACION TRIBUTARIA,OU=SAT-IES Authority,emailAddress=contacto.tecnic")
    # assert cert.subject().startswith("CN=SERVICIO DE ADMINISTRACION TRIBUTARIA,name=SERVICIO DE ADMINISTRACION TRIBUTARIA,O=SERVICIO DE ADMINISTRACION TRIBUTARIA,x500Unique")
    assert cert.certificate_base64().startswith("MIIGKTCCBBGgAwIBAgIUMDAwMDEwMDAwMDA1MDQ0NjUwM")
    assert str(cert.serial_number) == '275106190557734483187066766829380167334928331320'
    assert cert.certificate_number == "00001000000504465028"
    assert cert.branch_name == "MEGAPACSAT970701NN301"
    assert cert.rfc_pac == "SAT970701NN3"
    assert cert.rfc_representante == "GARJ7504168N4"
    assert cert.curp is None
    assert cert.curp_representante == "GARJ750416HDFRVR09"
    assert str(cert.serial_number) == "275106190557734483187066766829380167334928331320"
    assert cert.rfc_pac == "SAT970701NN3"
    assert cert.rfc_representante == "GARJ7504168N4"
    assert cert.legal_name == 'SERVICIO DE ADMINISTRACION TRIBUTARIA'
    assert cert.email is None


def test_signer_razon_social():
    res = {
        'CACX7605101P8': 'XOCHILT CASAS CHAVEZ',
        'FUNK671228PH6': 'KARLA FUENTE NOLASCO',
        'IAÑL750210963': 'LUIS IAN ÑUZCO',
        'JUFA7608212V6': 'ADRIANA JUAREZ FERNANDEZ',
        'KAHO641101B39': 'OSCAR KALA HAAK',
        'KICR630120NX3': 'RODRIGO KITIA CASTRO',
        'MISC491214B86': 'CECILIA MIRANDA SANCHEZ',
        'RAQÑ7701212M3': 'ÑEVES RAMIREZ QUEZADA',
        'WATM640917J45': 'MARIA WATEMBER TORRES',
        'WERX631016S30': 'XAIME WEIR ROJO',
        'XAMA620210DQ5': 'ALBA XKARAJAM MENDEZ',
        'XIQB891116QE4': 'BERENICE XIMO QUEZADA',
        'XOJI740919U48': 'INGRID XODAR JIMENEZ',
        'EKU9003173C9': 'ESCUELA KEMPER URGATE',
        'EWE1709045U0': 'ESCUELA WILSON ESQUIVEL',
        'H&E951128469': 'HERRERIA & ELECTRICOS',
        'HAÑ930228SM9': 'HERMANOS ANZURES ÑARVAEZ',
        'IIA040805DZ4': 'INDISTRIA ILUMINADORA DE ALMACENES',
        'IVD920810GU2': 'INNOVACION VALOR Y DESARROLLO',
        'IXS7607092R5': 'INTERNACIONAL XIMBO Y SABORES',
        'JES900109Q90': 'JIMENEZ ESTRADA SALAS A A',
        'KIJ0906199R1': 'KERNEL INDUSTIA JUGUETERA',
        'L&O950913MSA': 'LUCES & OBRAS',
        'OÑO120726RX3': 'ORGANICOS ÑAVEZ OSORIO S.A DE C.V',
        'S&S051221SE2': 'S & SOFTWARE',
        'URE180429TM6': 'UNIVERSIDAD ROBOTICA ESPAÑOLA',
        'XIA190128J61': 'XENON INDUSTRIAL ARTICLES',
        'ZUÑ920208KL4': 'ZAPATERIA URTADO ÑERI'
    }

    for rfc in chain(PERSONAS_FISICAS, PERSONAS_MORALES):
        try:
            signer = get_signer(rfc, get_csd=True)
            rs = signer.legal_name
            assert res[rfc] == rs
            assert rfc == signer.rfc
        except FileNotFoundError as ex:
            pass


def test_curp():
    curp = CURP("FEPM890204HASRRN08")
    assert curp.is_valid()
    assert curp.estado == "AS"

    # for rfc in PERSONAS_FISICAS:
    #     signer = get_signer(rfc)
    #     assert not signer.curp.is_valid()


def test_crypto_razon_social():
    for rfc in chain(PERSONAS_FISICAS, PERSONAS_MORALES):
        try:
            signer = get_signer(rfc, get_csd=True)

            cer = crypto.load_certificate(crypto.FILETYPE_ASN1, signer.certificate_bytes())
            subj = cer.get_subject()
            for k, v in subj.get_components():
                k.decode("latin_1")
                v.decode("latin_1")

        except (FileNotFoundError, ValueError) as ex:
            pass


def test_verify_certificates():
    for rfc in chain(PERSONAS_FISICAS, PERSONAS_MORALES):
        with mock.patch(f'{module}.transform.SAT_Certificate_Store', SAT_Certificate_Store_Pruebas):
            try:
                signer = get_signer(rfc)
                verify_certificate(signer, at=datetime(2021, 6, 12))
                assert signer.type == CertificateType.Fiel

                signer_csd = get_signer(rfc, get_csd=True)
                verify_certificate(signer_csd, at=datetime(2021, 6, 12))
                assert signer_csd.type == CertificateType.CSD

            except FileNotFoundError as ex:
                pass


def test_requirement():
    for rfc in PERSONAS_FISICAS:
        res = CertificateSigningRequest.load(
            request=open(os.path.join(current_dir, "csd", "Personas Fisicas", f'{rfc.lower()}.req'), 'rb').read()
        )
    for rfc in PERSONAS_MORALES:
        res = CertificateSigningRequest.load(
            request=open(os.path.join(current_dir, "csd", "Personas Morales", f'{rfc.lower()}.req'), 'rb').read()
        )


def test_thumbprint():
    signer = get_signer("xiqb891116qe4")

    assert signer.fingerprint().hex() == 'cad8b0aa7940aec665d9db9755b995b8631409c4'

    signer2 = get_signer("h&e951128469")

    with pytest.raises(CFDIError) as ex:
        signer = Signer(
            key=signer.key,
            certificate=signer2.certificate,
            check=True
        )

    password = "IGUESSTHISISSECURE"
    p12 = signer.pcks12_bytes(password)
    with open(os.path.join(current_dir, "csd", f'xiqb891116qe4.pfx'), 'wb') as f:
        f.write(p12)

    pkcs12_signer = Signer.load_pkcs12(
        data=open(os.path.join(current_dir, "csd", f'xiqb891116qe4.pfx'), 'rb').read(),
        password=password
    )

    assert pkcs12_signer.legal_name == 'BERENICE XIMO QUEZADA'

    os.remove(os.path.join(current_dir, "csd", f'xiqb891116qe4.pfx'))


def test_signer_change_password():
    rfc = 'xiqb891116qe4'
    signer = get_signer(rfc)

    password = "THISISMORESCURE"
    key = signer.key_bytes(password)
    cer = signer.certificate_bytes()

    with open(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.key'), 'wb') as f:
        f.write(key)
    with open(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.cer'), 'wb') as f:
        f.write(cer)
    with open(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.txt'), 'w') as f:
        f.write(password)

    signer = Signer.load(
        certificate=open(os.path.join(current_dir, f'csd/{rfc}.out.cer'), 'rb').read(),
        key=open(os.path.join(current_dir, f'csd/{rfc}.out.key'), 'rb').read(),
        password=open(os.path.join(current_dir, f'csd/{rfc}.out.txt'), 'rb').read()
    )

    os.remove(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.key'))
    os.remove(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.cer'))
    os.remove(os.path.join(current_dir, "csd", 'xiqb891116qe4.out.txt'))


def test_rfc():
    for rfc in PERSONAS_FISICAS:
        r = RFC(rfc, entity_type=RFCType.FISICA)
        assert r.is_valid()

    for rfc in PERSONAS_MORALES:
        r = RFC(rfc, entity_type=RFCType.MORAL)
        assert r.is_valid()

    r = RFC_Generico_Nacional
    assert r.is_valid()
    assert r.type == RFCType.FISICA

    r = RFC_Generico_Extranjero
    assert r.is_valid()
    assert r.type == RFCType.FISICA


def test_period():
    period = DatePeriod(year=2020, month=4)
    my_date = date(year=2021, month=3, day=20)
    assert my_date > period
    assert my_date >= period
    assert my_date != period
    assert not my_date == period
    assert not my_date <= period
    assert not my_date < period

    assert str(period) == '2020-04'

    period = DatePeriod(year=None)
    my_date = date(year=2021, month=3, day=20)
    assert my_date >= period
    assert my_date == period
    assert my_date <= period
    assert not my_date > period
    assert not my_date < period
    assert not my_date != period

    assert str(period) == ""

    assert not period == 1
    assert period != 1
    with pytest.raises(TypeError):
        assert not period < 1
    with pytest.raises(TypeError):
        assert not period > 1
    with pytest.raises(TypeError):
        assert not period >= 1
    with pytest.raises(TypeError):
        assert not period <= 1


def test_code():
    class Salutaciones(StrEnum):
        Hola = '01'
        Adios = '02'

    c = Code('01', 'Hola')

    assert Salutaciones.Hola == c
    assert c == Salutaciones.Hola
    assert c == '01'
    assert '01' == c
    assert c == Code('01', 'OtherHola')
